"""
Extracts a list of fablabs in the world, restrict to the ones in france, then format it both for a nice console output
and a flat txt file.

.. graphviz::

    digraph {
        rankdir = LR;
        stylesheet = "../_static/graphs.css";

        BEGIN [shape="point"];
        BEGIN -> "ODS()" -> "normalize" -> "filter_france" -> "Tee()" -> "JsonWriter()";
    }

"""

import json

from colorama import Fore, Style

import bonobo
from bonobo.commands.run import get_default_services
from bonobo.ext.opendatasoft import OpenDataSoftAPI

try:
    import pycountry
except ImportError as exc:
    raise ImportError(
        'You must install package "pycountry" to run this example.'
    ) from exc

API_DATASET = 'fablabs-in-the-world'
API_NETLOC = 'datanova.laposte.fr'
ROWS = 100


def _getlink(x):
    return x.get('url', None)


def normalize(row):
    result = {
        **
        row,
        'links': list(filter(None, map(_getlink, json.loads(row.get('links'))))),
        'country': pycountry.countries.get(alpha_2=row.get('country_code', '').upper()).name,
    }
    return result


def display(row):
    print(Style.BRIGHT, row.get('name'), Style.RESET_ALL, sep='')

    address = list(
        filter(
            None, (
                ' '.join(
                    filter(
                        None, (
                            row.get('postal_code', None),
                            row.get('city', None)
                        )
                    )
                ),
                row.get('county', None),
                row.get('country'),
            )
        )
    )

    print(
        '  - {}address{}: {address}'.format(
            Fore.BLUE, Style.RESET_ALL, address=', '.join(address)
        )
    )
    print(
        '  - {}links{}: {links}'.format(
            Fore.BLUE, Style.RESET_ALL, links=', '.join(row['links'])
        )
    )
    print(
        '  - {}geometry{}: {geometry}'.format(
            Fore.BLUE, Style.RESET_ALL, **row
        )
    )
    print(
        '  - {}source{}: {source}'.format(
            Fore.BLUE, Style.RESET_ALL, source='datanova/' + API_DATASET
        )
    )


graph = bonobo.Graph(
    OpenDataSoftAPI(
        dataset=API_DATASET, netloc=API_NETLOC, timezone='Europe/Paris'
    ),
    normalize,
    bonobo.Filter(filter=lambda row: row.get('country') == 'France'),
    bonobo.JsonWriter(path='fablabs.txt', ioformat='arg0'),
    bonobo.Tee(display),
)

if __name__ == '__main__':
    bonobo.run(graph, services=get_default_services(__file__))



from bonobo.config.processors import ContextProcessor
from bonobo.constants import BEGIN, END
from bonobo.execution.graph import GraphExecutionContext
from bonobo.strategies import NaiveStrategy
from bonobo.structs import Bag, Graph


def generate_integers():
    yield from range(10)


def square(i: int) -> int:
    return i**2


def push_result(results, i: int):
    results.append(i)


@ContextProcessor.decorate(push_result)
def results(f, context):
    results = []
    yield results
    context.parent.results = results


chain = (generate_integers, square, push_result)


def test_empty_execution_context():
    graph = Graph()

    ctx = GraphExecutionContext(graph)
    assert not len(ctx.nodes)
    assert not len(ctx.plugins)

    assert not ctx.alive


def test_execution():
    graph = Graph()
    graph.add_chain(*chain)

    strategy = NaiveStrategy()
    ctx = strategy.execute(graph)

    assert ctx.results == [1, 4, 9, 16, 25, 36, 49, 64, 81]


def test_simple_execution_context():
    graph = Graph()
    graph.add_chain(*chain)

    ctx = GraphExecutionContext(graph)
    assert len(ctx.nodes) == len(chain)
    assert not len(ctx.plugins)

    for i, node in enumerate(chain):
        assert ctx[i].wrapped is node

    assert not ctx.alive
    assert not ctx.started
    assert not ctx.stopped

    ctx.write(BEGIN, Bag(), END)

    assert not ctx.alive
    assert not ctx.started
    assert not ctx.stopped

    ctx.start()

    assert ctx.alive
    assert ctx.started
    assert not ctx.stopped

    ctx.stop()

    assert not ctx.alive
    assert ctx.started
    assert ctx.stopped



""" Iterator utilities. """
import functools


def force_iterator(mixed):
    """Sudo make me an iterator.

    Deprecated?

    :param mixed:
    :return: Iterator, baby.
    """
    if isinstance(mixed, str):
        return [mixed]
    try:
        return iter(mixed)
    except TypeError:
        return [mixed] if mixed else []


def ensure_tuple(tuple_or_mixed):
    if isinstance(tuple_or_mixed, tuple):
        return tuple_or_mixed
    return (tuple_or_mixed, )


def tuplize(generator):
    """ Takes a generator and make it a tuple-returning function. As a side
    effect, it can also decorate any iterator-returning function to force
    return value to be a tuple.
    """

    @functools.wraps(generator)
    def tuplized(*args, **kwargs):
        return tuple(generator(*args, **kwargs))

    return tuplized


def iter_if_not_sequence(mixed):
    if isinstance(mixed, (
        dict,
        list,
        str,
        bytes,
    )):
        raise TypeError(type(mixed).__name__)
    return iter(mixed)



import logging

from bonobo.ext.jupyter.widget import BonoboWidget
from bonobo.plugins import Plugin

try:
    import IPython.core.display
except ImportError as e:
    logging.exception(
        'You must install Jupyter to use the bonobo Jupyter extension. Easiest way is to install the '
        'optional "jupyter" dependencies with «pip install bonobo[jupyter]», but you can also install a '
        'specific version by yourself.'
    )


class JupyterOutputPlugin(Plugin):
    def initialize(self):
        self.widget = BonoboWidget()
        IPython.core.display.display(self.widget)

    def run(self):
        self.widget.value = [
            str(self.context.parent[i]) for i in self.context.parent.graph.topologically_sorted_indexes
        ]

    finalize = run



'''
This example shows how a different file system service can be injected
into a transformation (as compressing pickled objects often makes sense
anyways).  The pickle itself contains a list of lists as follows:

```
[
    ['category', 'sms'],
    ['ham', 'Go until jurong point, crazy..'],
    ['ham', 'Ok lar... Joking wif u oni...'],
    ['spam', 'Free entry in 2 a wkly comp to win...'],
    ['ham', 'U dun say so early hor... U c already then say...'],
    ['ham', 'Nah I don't think he goes to usf, he lives around here though'],
    ['spam', 'FreeMsg Hey there darling it's been 3 week's now...'],
    ...
]
```

where the first column categorizes and sms as "ham" or "spam".  The second
column contains the sms itself.

Data set taken from:
https://www.kaggle.com/uciml/sms-spam-collection-dataset/downloads/sms-spam-collection-dataset.zip

The transformation (1) reads the pickled data, (2) marks and shortens
messages categorized as spam, and (3) prints the output.

'''

import bonobo
from bonobo.commands.run import get_default_services
from fs.tarfs import TarFS


def cleanse_sms(**row):
    if row['category'] == 'spam':
        row['sms_clean'] = '**MARKED AS SPAM** ' + row['sms'][0:50] + (
            '...' if len(row['sms']) > 50 else ''
        )
    else:
        row['sms_clean'] = row['sms']

    return row['sms_clean']


graph = bonobo.Graph(
    # spam.pkl is within the gzipped tarball
    bonobo.PickleReader('spam.pkl'),
    cleanse_sms,
    bonobo.PrettyPrinter(),
)


def get_services():
    return {'fs': TarFS(bonobo.get_examples_path('datasets/spam.tgz'))}


if __name__ == '__main__':
    bonobo.run(graph, services=get_default_services(__file__))



from bonobo import settings
from bonobo.config import Configurable, ContextProcessor, Option, Service
from bonobo.errors import UnrecoverableValueError, UnrecoverableNotImplementedError
from bonobo.structs.bags import Bag


class IOFormatEnabled(Configurable):
    ioformat = Option(default=settings.IOFORMAT.get)

    def get_input(self, *args, **kwargs):
        if self.ioformat == settings.IOFORMAT_ARG0:
            if len(args) != 1 or len(kwargs):
                raise UnrecoverableValueError(
                    'Wrong input formating: IOFORMAT=ARG0 implies one arg and no kwargs, got args={!r} and kwargs={!r}.'.
                    format(args, kwargs)
                )
            return args[0]

        if self.ioformat == settings.IOFORMAT_KWARGS:
            if len(args) or not len(kwargs):
                raise UnrecoverableValueError(
                    'Wrong input formating: IOFORMAT=KWARGS ioformat implies no arg, got args={!r} and kwargs={!r}.'.
                    format(args, kwargs)
                )
            return kwargs

        raise UnrecoverableNotImplementedError('Unsupported format.')

    def get_output(self, row):
        if self.ioformat == settings.IOFORMAT_ARG0:
            return row

        if self.ioformat == settings.IOFORMAT_KWARGS:
            return Bag(**row)

        raise UnrecoverableNotImplementedError('Unsupported format.')


class FileHandler(Configurable):
    """Abstract component factory for file-related components.

    Args:
        path (str): which path to use within the provided filesystem.
        eol (str): which character to use to separate lines.
        mode (str): which mode to use when opening the file.
        fs (str): service name to use for filesystem.
    """

    path = Option(str, required=True, positional=True)  # type: str
    eol = Option(str, default='\n')  # type: str
    mode = Option(str)  # type: str
    encoding = Option(str, default='utf-8')  # type: str
    fs = Service('fs')  # type: str

    @ContextProcessor
    def file(self, context, fs):
        with self.open(fs) as file:
            yield file

    def open(self, fs):
        return fs.open(self.path, self.mode, encoding=self.encoding)


class Reader:
    """Abstract component factory for readers.
    """

    def __call__(self, *args, **kwargs):
        yield from self.read(*args, **kwargs)

    def read(self, *args, **kwargs):
        raise NotImplementedError('Abstract.')


class Writer:
    """Abstract component factory for writers.
    """

    def __call__(self, *args, **kwargs):
        return self.write(*args, **kwargs)

    def write(self, *args, **kwargs):
        raise NotImplementedError('Abstract.')



from collections import Iterable
from contextlib import contextmanager

from bonobo.config.options import Option
from bonobo.util.compat import deprecated_alias
from bonobo.util.iterators import ensure_tuple

_CONTEXT_PROCESSORS_ATTR = '__processors__'


class ContextProcessor(Option):
    """
    A ContextProcessor is a kind of transformation decorator that can setup and teardown a transformation and runtime
    related dependencies, at the execution level.

    It works like a yielding context manager, and is the recommended way to setup and teardown objects you'll need
    in the context of one execution. It's the way to overcome the stateless nature of transformations.

    The yielded values will be passed as positional arguments to the next context processors (order do matter), and
    finally to the __call__ method of the transformation.

    Warning: this may change for a similar but simpler implementation, don't relly too much on it (yet).

    Example:

        >>> from bonobo.config import Configurable
        >>> from bonobo.util.objects import ValueHolder

        >>> class Counter(Configurable):
        ...     @ContextProcessor
        ...     def counter(self, context):
        ...         yield ValueHolder(0)
        ...
        ...     def __call__(self, counter, *args, **kwargs):
        ...         counter += 1
        ...         yield counter.get()

    """

    @property
    def __name__(self):
        return self.func.__name__

    def __init__(self, func):
        self.func = func
        super(ContextProcessor, self).__init__(required=False, default=self.__name__)
        self.name = self.__name__

    def __repr__(self):
        return repr(self.func).replace('<function', '<{}'.format(type(self).__name__))

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    @classmethod
    def decorate(cls, cls_or_func):
        try:
            cls_or_func.__processors__
        except AttributeError:
            cls_or_func.__processors__ = []

        def decorator(processor, cls_or_func=cls_or_func):
            cls_or_func.__processors__.append(cls(processor))
            return cls_or_func

        return decorator


class ContextCurrifier:
    """
    This is a helper to resolve processors.
    """

    def __init__(self, wrapped, *initial_context):
        self.wrapped = wrapped
        self.context = tuple(initial_context)
        self._stack, self._stack_values = None, None

    def __iter__(self):
        yield from self.wrapped

    def __call__(self, *args, **kwargs):
        if not callable(self.wrapped) and isinstance(self.wrapped, Iterable):
            return self.__iter__()
        return self.wrapped(*self.context, *args, **kwargs)

    def setup(self, *context):
        if self._stack is not None:
            raise RuntimeError('Cannot setup context currification twice.')

        self._stack, self._stack_values = list(), list()
        for processor in resolve_processors(self.wrapped):
            _processed = processor(self.wrapped, *context, *self.context)
            _append_to_context = next(_processed)
            self._stack_values.append(_append_to_context)
            if _append_to_context is not None:
                self.context += ensure_tuple(_append_to_context)
            self._stack.append(_processed)

    def teardown(self):
        while self._stack:
            processor = self._stack.pop()
            try:
                # todo yield from ? how to ?
                processor.send(self._stack_values.pop())
            except StopIteration as exc:
                # This is normal, and wanted.
                pass
            else:
                # No error ? We should have had StopIteration ...
                raise RuntimeError('Context processors should not yield more than once.')
        self._stack, self._stack_values = None, None

    @contextmanager
    def as_contextmanager(self, *context):
        """
        Convenience method to use it as a contextmanager, mostly for test purposes.

        Example:

            >>> with ContextCurrifier(node).as_contextmanager(context) as stack:
            ...     stack()

        :param context:
        :return:
        """
        self.setup(*context)
        yield self
        self.teardown()


def resolve_processors(mixed):
    try:
        yield from mixed.__processors__
    except AttributeError:
        yield from ()


get_context_processors = deprecated_alias('get_context_processors', resolve_processors)



from copy import copy

from bonobo.constants import BEGIN


class Graph:
    """
    Represents a directed graph of nodes.
    """

    def __init__(self, *chain):
        self.edges = {BEGIN: set()}
        self.named = {}
        self.nodes = []
        self.add_chain(*chain)

    def __iter__(self):
        yield from self.nodes

    def __len__(self):
        """ Node count.
        """
        return len(self.nodes)

    def __getitem__(self, key):
        return self.nodes[key]

    def outputs_of(self, idx, create=False):
        """ Get a set of the outputs for a given node index.
        """
        if create and not idx in self.edges:
            self.edges[idx] = set()
        return self.edges[idx]

    def add_node(self, c):
        """ Add a node without connections in this graph and returns its index.
        """
        idx = len(self.nodes)
        self.edges[idx] = set()
        self.nodes.append(c)
        return idx

    def add_chain(self, *nodes, _input=BEGIN, _output=None, _name=None):
        """ Add a chain in this graph.
        """
        if len(nodes):
            _input = self._resolve_index(_input)
            _output = self._resolve_index(_output)

            for i, node in enumerate(nodes):
                _next = self.add_node(node)
                if not i and _name:
                    if _name in self.named:
                        raise KeyError('Duplicate name {!r} in graph.'.format(_name))
                    self.named[_name] = _next
                self.outputs_of(_input, create=True).add(_next)
                _input = _next

            if _output is not None:
                self.outputs_of(_input, create=True).add(_output)

            if hasattr(self, '_topologcally_sorted_indexes_cache'):
                del self._topologcally_sorted_indexes_cache

        return self

    def copy(self):
        g = Graph()

        g.edges = copy(self.edges)
        g.named = copy(self.named)
        g.nodes = copy(self.nodes)

        return g

    @property
    def topologically_sorted_indexes(self):
        """Iterate in topological order, based on networkx's topological_sort() function.
        """
        try:
            return self._topologcally_sorted_indexes_cache
        except AttributeError:
            seen = set()
            order = []
            explored = set()

            for i in self.edges:
                if i in explored:
                    continue
                fringe = [i]
                while fringe:
                    w = fringe[-1]  # depth first search
                    if w in explored:  # already looked down this branch
                        fringe.pop()
                        continue
                    seen.add(w)  # mark as seen
                    # Check successors for cycles and for new nodes
                    new_nodes = []
                    for n in self.outputs_of(w):
                        if n not in explored:
                            if n in seen:  # CYCLE !!
                                raise RuntimeError("Graph contains a cycle.")
                            new_nodes.append(n)
                    if new_nodes:  # Add new_nodes to fringe
                        fringe.extend(new_nodes)
                    else:  # No new nodes so w is fully explored
                        explored.add(w)
                        order.append(w)
                        fringe.pop()  # done considering this node
            self._topologcally_sorted_indexes_cache = tuple(filter(lambda i: type(i) is int, reversed(order)))
            return self._topologcally_sorted_indexes_cache

    def _resolve_index(self, mixed):
        """ Find the index based on various strategies for a node, probably an input or output of chain. Supported inputs are indexes, node values or names.
        """
        if mixed is None:
            return None

        if type(mixed) is int or mixed in self.edges:
            return mixed

        if isinstance(mixed, str) and mixed in self.named:
            return self.named[mixed]

        if mixed in self.nodes:
            return self.nodes.index(mixed)

        raise ValueError('Cannot find node matching {!r}.'.format(mixed))



import sys
from textwrap import indent


def _get_error_message(exc):
    if hasattr(exc, '__str__'):
        message = str(exc)
        return message[0].upper() + message[1:]
    return '\n'.join(exc.args),


def print_error(exc, trace, context=None, method=None):
    """
    Error handler. Whatever happens in a plugin or component, if it looks like an exception, taste like an exception
    or somehow make me think it is an exception, I'll handle it.

    :param exc: the culprit
    :param trace: Hercule Poirot's logbook.
    :return: to hell
    """

    from colorama import Fore, Style

    prefix = '{}{} | {}'.format(Fore.RED, Style.BRIGHT, Style.RESET_ALL)

    print(
        Style.BRIGHT,
        Fore.RED,
        type(exc).__name__,
        ' (in {}{})'.format(type(context).__name__, '.{}()'.format(method) if method else '') if context else '',
        Style.RESET_ALL,
        '\n',
        indent(_get_error_message(exc), prefix + Style.BRIGHT),
        Style.RESET_ALL,
        sep='',
        file=sys.stderr,
    )
    print(prefix, file=sys.stderr)
    print(indent(trace, prefix, predicate=lambda line: True), file=sys.stderr)



import bonobo
from bonobo.commands.run import get_default_services
from bonobo.nodes.factory import Factory
from bonobo.nodes.io.json import JsonDictItemsReader

normalize = Factory()
normalize[0].str().title()
normalize.move(0, 'title')
normalize.move(0, 'address')

graph = bonobo.Graph(
    JsonDictItemsReader('datasets/coffeeshops.json'),
    normalize,
    bonobo.PrettyPrinter(),
)

if __name__ == '__main__':
    bonobo.run(graph, services=get_default_services(__file__))



"""
Example on how to use symple python dictionaries to communicate between transformations.

.. graphviz::

    digraph {
        rankdir = LR;
        stylesheet = "../_static/graphs.css";

        BEGIN [shape="point"];
        BEGIN -> "extract()" -> "transform(row: dict)" -> "load(row: dict)";
    }

"""

from random import randint

from bonobo import Graph


def extract():
    yield {'topic': 'foo'}
    yield {'topic': 'bar'}
    yield {'topic': 'baz'}


def transform(row: dict):
    return {
        'topic': row['topic'].title(),
        'randint': randint(10, 99),
    }


def load(row: dict):
    print(row)


graph = Graph(extract, transform, load)

if __name__ == '__main__':
    from bonobo import run

    run(graph)



"""
Simple example of :func:`bonobo.count` usage.

.. graphviz::

    digraph {
        rankdir = LR;
        stylesheet = "../_static/graphs.css";

        BEGIN [shape="point"];
        BEGIN -> "range()" -> "count" -> "print";
    }

"""

import bonobo

graph = bonobo.Graph(range(42), bonobo.count, print)

if __name__ == '__main__':
    bonobo.run(graph)



import functools
import itertools

from bonobo import settings
from bonobo.config import Configurable, Option
from bonobo.config.processors import ContextProcessor
from bonobo.structs.bags import Bag
from bonobo.util.objects import ValueHolder
from bonobo.util.term import CLEAR_EOL

from bonobo.constants import NOT_MODIFIED

__all__ = [
    'Limit',
    'PrettyPrinter',
    'Tee',
    'arg0_to_kwargs',
    'count',
    'identity',
    'kwargs_to_arg0',
    'noop',
]


def identity(x):
    return x


class Limit(Configurable):
    """
    Creates a Limit() node, that will only let go through the first n rows (defined by the `limit` option), unmodified.

    .. attribute:: limit

        Number of rows to let go through.

    """
    limit = Option(positional=True, default=10)

    @ContextProcessor
    def counter(self, context):
        yield ValueHolder(0)

    def call(self, counter, *args, **kwargs):
        counter += 1
        if counter <= self.limit:
            yield NOT_MODIFIED


def Tee(f):
    from bonobo.constants import NOT_MODIFIED

    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        nonlocal f
        f(*args, **kwargs)
        return NOT_MODIFIED

    return wrapped


def count(counter, *args, **kwargs):
    counter += 1


@ContextProcessor.decorate(count)
def _count_counter(self, context):
    counter = ValueHolder(0)
    yield counter
    context.send(Bag(counter._value))


class PrettyPrinter(Configurable):
    def call(self, *args, **kwargs):
        formater = self._format_quiet if settings.QUIET.get() else self._format_console

        for i, (item, value) in enumerate(itertools.chain(enumerate(args), kwargs.items())):
            print(formater(i, item, value))

    def _format_quiet(self, i, item, value):
        return ' '.join(((' ' if i else '-'), str(item), ':', str(value).strip()))

    def _format_console(self, i, item, value):
        return ' '.join(
            ((' ' if i else '•'), str(item), '=', str(value).strip().replace('\n', '\n' + CLEAR_EOL), CLEAR_EOL)
        )


def noop(*args, **kwargs):  # pylint: disable=unused-argument
    from bonobo.constants import NOT_MODIFIED
    return NOT_MODIFIED


def arg0_to_kwargs(row):
    """
    Transform items in a stream from "arg0" format (each call only has one positional argument, which is a dict-like
    object) to "kwargs" format (each call only has keyword arguments that represent a row).

    :param row:
    :return: bonobo.Bag
    """
    return Bag(**row)


def kwargs_to_arg0(**row):
    """
    Transform items in a stream from "kwargs" format (each call only has keyword arguments that represent a row) to
    "arg0" format (each call only has one positional argument, which is a dict-like object) .

    :param **row:
    :return: bonobo.Bag
    """
    return Bag(row)



import logging

from bonobo.nodes import CsvReader, CsvWriter, FileReader, FileWriter, Filter, JsonReader, JsonWriter, Limit, \
    PickleReader, PickleWriter, PrettyPrinter, RateLimited, Tee, arg0_to_kwargs, count, identity, kwargs_to_arg0, noop
from bonobo.strategies import create_strategy
from bonobo.structs import Bag, ErrorBag, Graph, Token
from bonobo.util import get_name

__all__ = []


def register_api(x, __all__=__all__):
    __all__.append(get_name(x))
    return x


def register_api_group(*args):
    for attr in args:
        register_api(attr)


@register_api
def run(graph, strategy=None, plugins=None, services=None):
    """
    Main entry point of bonobo. It takes a graph and creates all the necessary plumbery around to execute it.

    The only necessary argument is a :class:`Graph` instance, containing the logic you actually want to execute.

    By default, this graph will be executed using the "threadpool" strategy: each graph node will be wrapped in a
    thread, and executed in a loop until there is no more input to this node.

    You can provide plugins factory objects in the plugins list, this function will add the necessary plugins for
    interactive console execution and jupyter notebook execution if it detects correctly that it runs in this context.

    You'll probably want to provide a services dictionary mapping service names to service instances.

    :param Graph graph: The :class:`Graph` to execute.
    :param str strategy: The :class:`bonobo.strategies.base.Strategy` to use.
    :param list plugins: The list of plugins to enhance execution.
    :param dict services: The implementations of services this graph will use.
    :return bonobo.execution.graph.GraphExecutionContext:
    """
    strategy = create_strategy(strategy)

    plugins = plugins or []

    from bonobo import settings
    settings.check()

    if not settings.QUIET.get():  # pragma: no cover
        if _is_interactive_console():
            from bonobo.ext.console import ConsoleOutputPlugin
            if ConsoleOutputPlugin not in plugins:
                plugins.append(ConsoleOutputPlugin)

        if _is_jupyter_notebook():
            try:
                from bonobo.ext.jupyter import JupyterOutputPlugin
            except ImportError:
                logging.warning(
                    'Failed to load jupyter widget. Easiest way is to install the optional "jupyter" '
                    'dependencies with «pip install bonobo[jupyter]», but you can also install a specific '
                    'version by yourself.'
                )
            else:
                if JupyterOutputPlugin not in plugins:
                    plugins.append(JupyterOutputPlugin)

    return strategy.execute(graph, plugins=plugins, services=services)


# bonobo.structs
register_api_group(Bag, ErrorBag, Graph, Token)

# bonobo.strategies
register_api(create_strategy)


# Shortcut to filesystem2's open_fs, that we make available there for convenience.
@register_api
def open_fs(fs_url=None, *args, **kwargs):
    """
    Wraps :func:`fs.open_fs` function with a few candies.

    :param str fs_url: A filesystem URL
    :param parse_result: A parsed filesystem URL.
    :type parse_result: :class:`ParseResult`
    :param bool writeable: True if the filesystem must be writeable.
    :param bool create: True if the filesystem should be created if it does not exist.
    :param str cwd: The current working directory (generally only relevant for OS filesystems).
    :param str default_protocol: The protocol to use if one is not supplied in the FS URL (defaults to ``"osfs"``).
    :returns: :class:`~fs.base.FS` object
    """
    from fs import open_fs as _open_fs
    from os.path import expanduser
    from os import getcwd

    if fs_url is None:
        fs_url = getcwd()

    return _open_fs(expanduser(str(fs_url)), *args, **kwargs)


# bonobo.nodes
register_api_group(
    CsvReader,
    CsvWriter,
    FileReader,
    FileWriter,
    Filter,
    JsonReader,
    JsonWriter,
    Limit,
    PickleReader,
    PickleWriter,
    PrettyPrinter,
    RateLimited,
    Tee,
    arg0_to_kwargs,
    count,
    identity,
    kwargs_to_arg0,
    noop,
)


def _is_interactive_console():
    import sys
    return sys.stdout.isatty()


def _is_jupyter_notebook():
    try:
        return get_ipython().__class__.__name__ == 'ZMQInteractiveShell'
    except NameError:
        return False


@register_api
def get_examples_path(*pathsegments):
    import os
    import pathlib
    return str(pathlib.Path(os.path.dirname(__file__), 'examples', *pathsegments))


@register_api
def open_examples_fs(*pathsegments):
    return open_fs(get_examples_path(*pathsegments))



import os
import runpy
import sys
from unittest.mock import patch

import pkg_resources
import pytest

from bonobo import __main__, __version__, get_examples_path
from bonobo.commands import entrypoint


def runner_entrypoint(*args):
    """ Run bonobo using the python command entrypoint directly (bonobo.commands.entrypoint). """
    return entrypoint(list(args))


def runner_module(*args):
    """ Run bonobo using the bonobo.__main__ file, which is equivalent as doing "python -m bonobo ..."."""
    with patch.object(sys, 'argv', ['bonobo', *args]):
        return runpy.run_path(__main__.__file__, run_name='__main__')


all_runners = pytest.mark.parametrize('runner', [runner_entrypoint, runner_module])


def test_entrypoint():
    commands = {}

    for command in pkg_resources.iter_entry_points('bonobo.commands'):
        commands[command.name] = command

    assert 'init' in commands
    assert 'run' in commands
    assert 'version' in commands


@all_runners
def test_no_command(runner, capsys):
    with pytest.raises(SystemExit):
        runner()
    _, err = capsys.readouterr()
    assert 'error: the following arguments are required: command' in err


@all_runners
def test_run(runner, capsys):
    runner('run', '--quiet', get_examples_path('types/strings.py'))
    out, err = capsys.readouterr()
    out = out.split('\n')
    assert out[0].startswith('Foo ')
    assert out[1].startswith('Bar ')
    assert out[2].startswith('Baz ')


@all_runners
def test_run_module(runner, capsys):
    runner('run', '--quiet', '-m', 'bonobo.examples.types.strings')
    out, err = capsys.readouterr()
    out = out.split('\n')
    assert out[0].startswith('Foo ')
    assert out[1].startswith('Bar ')
    assert out[2].startswith('Baz ')


@all_runners
def test_run_path(runner, capsys):
    runner('run', '--quiet', get_examples_path('types'))
    out, err = capsys.readouterr()
    out = out.split('\n')
    assert out[0].startswith('Foo ')
    assert out[1].startswith('Bar ')
    assert out[2].startswith('Baz ')


@all_runners
def test_install_requirements_for_dir(runner):
    dirname = get_examples_path('types')
    with patch('bonobo.commands.run._install_requirements') as install_mock:
        runner('run', '--install', dirname)
    install_mock.assert_called_once_with(os.path.join(dirname, 'requirements.txt'))


@all_runners
def test_install_requirements_for_file(runner):
    dirname = get_examples_path('types')
    with patch('bonobo.commands.run._install_requirements') as install_mock:
        runner('run', '--install', os.path.join(dirname, 'strings.py'))
    install_mock.assert_called_once_with(os.path.join(dirname, 'requirements.txt'))


@all_runners
def test_version(runner, capsys):
    runner('version')
    out, err = capsys.readouterr()
    out = out.strip()
    assert out.startswith('bonobo ')
    assert __version__ in out


@all_runners
def test_run_with_env(runner, capsys):
    runner(
        'run', '--quiet',
        get_examples_path('env_vars/get_passed_env.py'), '--env', 'ENV_TEST_NUMBER=123', '--env',
        'ENV_TEST_USER=cwandrews', '--env', "ENV_TEST_STRING='my_test_string'"
    )
    out, err = capsys.readouterr()
    out = out.split('\n')
    assert out[0] == 'cwandrews'
    assert out[1] == '123'
    assert out[2] == 'my_test_string'


@all_runners
def test_run_module_with_env(runner, capsys):
    runner(
        'run', '--quiet', '-m', 'bonobo.examples.env_vars.get_passed_env', '--env', 'ENV_TEST_NUMBER=123', '--env',
        'ENV_TEST_USER=cwandrews', '--env', "ENV_TEST_STRING='my_test_string'"
    )
    out, err = capsys.readouterr()
    out = out.split('\n')
    assert out[0] == 'cwandrews'
    assert out[1] == '123'
    assert out[2] == 'my_test_string'



from bonobo.util.collections import sortedlist
from bonobo.util.inspect import (
    inspect_node,
    isbag,
    isconfigurable,
    isconfigurabletype,
    iscontextprocessor,
    iserrorbag,
    isloopbackbag,
    ismethod,
    isoption,
    istype,
)
from bonobo.util.objects import (get_name, get_attribute_or_create, ValueHolder)
from bonobo.util.python import require

# Bonobo's util API
__all__ = [
    'ValueHolder',
    'get_attribute_or_create',
    'get_name',
    'inspect_node',
    'isbag',
    'isconfigurable',
    'isconfigurabletype',
    'iscontextprocessor',
    'iserrorbag',
    'isloopbackbag',
    'ismethod',
    'isoption',
    'istype',
    'require',
]



import traceback
from contextlib import contextmanager
from time import sleep

from bonobo.config import create_container
from bonobo.config.processors import ContextCurrifier
from bonobo.plugins import get_enhancers
from bonobo.util.errors import print_error
from bonobo.util.objects import Wrapper, get_name


@contextmanager
def recoverable(error_handler):
    try:
        yield
    except Exception as exc:  # pylint: disable=broad-except
        error_handler(exc, traceback.format_exc())


@contextmanager
def unrecoverable(error_handler):
    try:
        yield
    except Exception as exc:  # pylint: disable=broad-except
        error_handler(exc, traceback.format_exc())
        raise  # raise unrecoverableerror from x ?


class LoopingExecutionContext(Wrapper):
    alive = True
    PERIOD = 0.25

    @property
    def started(self):
        return self._started

    @property
    def stopped(self):
        return self._stopped

    def __init__(self, wrapped, parent, services=None):
        super().__init__(wrapped)

        self.parent = parent

        if services:
            if parent:
                raise RuntimeError(
                    'Having services defined both in GraphExecutionContext and child NodeExecutionContext is not supported, for now.'
                )
            self.services = create_container(services)
        else:
            self.services = None

        self._started, self._stopped = False, False
        self._stack = None

        # XXX enhancers
        self._enhancers = get_enhancers(self.wrapped)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        self.stop()

    def start(self):
        if self.started:
            raise RuntimeError('Cannot start a node twice ({}).'.format(get_name(self)))

        self._started = True

        self._stack = ContextCurrifier(self.wrapped, *self._get_initial_context())
        self._stack.setup(self)

        for enhancer in self._enhancers:
            with unrecoverable(self.handle_error):
                enhancer.start(self)

    def loop(self):
        """Generic loop. A bit boring. """
        while self.alive:
            self.step()
            sleep(self.PERIOD)

    def step(self):
        """Left as an exercise for the children."""
        raise NotImplementedError('Abstract.')

    def stop(self):
        if not self.started:
            raise RuntimeError('Cannot stop an unstarted node ({}).'.format(get_name(self)))

        if self._stopped:
            return

        try:
            if self._stack:
                self._stack.teardown()
        finally:
            self._stopped = True

    def handle_error(self, exc, trace):
        return print_error(exc, trace, context=self.wrapped)

    def _get_initial_context(self):
        if self.parent:
            return self.parent.services.args_for(self.wrapped)
        if self.services:
            return self.services.args_for(self.wrapped)
        return ()



def format_version(mod, *, name=None, quiet=False):
    from bonobo.util.pkgs import bonobo_packages
    args = {
        'name': name or mod.__name__,
        'version': mod.__version__,
        'location': bonobo_packages[name or mod.__name__].location
    }

    if not quiet:
        return '{name} v.{version} (in {location})'.format(**args)
    if quiet < 2:
        return '{name} {version}'.format(**args)
    if quiet < 3:
        return '{version}'.format(**args)

    raise RuntimeError('Hard to be so quiet...')


def execute(all=False, quiet=False):
    import bonobo
    from bonobo.util.pkgs import bonobo_packages

    print(format_version(bonobo, quiet=quiet))
    if all:
        for name in sorted(bonobo_packages):
            if name != 'bonobo':
                try:
                    mod = __import__(name.replace('-', '_'))
                    try:
                        print(format_version(mod, name=name, quiet=quiet))
                    except Exception as exc:
                        print('{} ({})'.format(name, exc))
                except ImportError as exc:
                    print('{} is not importable ({}).'.format(name, exc))


def register(parser):
    parser.add_argument('--all', '-a', action='store_true')
    parser.add_argument('--quiet', '-q', action='count')
    return execute



import time


class Timer:
    """
    Context manager used to time execution of stuff.
    """

    def __enter__(self):
        self.__start = time.time()

    def __exit__(self, type=None, value=None, traceback=None):
        # Error handling here
        self.__finish = time.time()

    @property
    def duration(self):
        return self.__finish - self.__start

    def __str__(self):
        return str(int(self.duration * 1000) / 1000.0) + 's'



class Token:
    """Factory for signal oriented queue messages or other token types."""

    def __init__(self, name):
        self.__name__ = name

    def __repr__(self):
        return '<{}>'.format(self.__name__)



# -*- coding: utf-8 -*-
#
# copyright 2012-2014 romain dorgueil
#
# licensed under the apache license, version 2.0 (the "license");
# you may not use this file except in compliance with the license.
# you may obtain a copy of the license at
#
#     http://www.apache.org/licenses/license-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the license is distributed on an "as is" basis,
# without warranties or conditions of any kind, either express or implied.
# see the license for the specific language governing permissions and
# limitations under the license.


class WithStatistics:
    def __init__(self, *names):
        self.statistics_names = names
        self.statistics = {name: 0 for name in names}

    def get_statistics(self, *args, **kwargs):
        return ((name, self.statistics[name]) for name in self.statistics_names)

    def get_statistics_as_string(self, *args, **kwargs):
        stats = tuple('{0}={1}'.format(name, cnt) for name, cnt in self.get_statistics(*args, **kwargs) if cnt > 0)
        return (kwargs.get('prefix', '') + ' '.join(stats)) if len(stats) else ''

    def increment(self, name):
        self.statistics[name] += 1



import os

import bonobo
from bonobo.constants import DEFAULT_SERVICES_ATTR, DEFAULT_SERVICES_FILENAME

DEFAULT_GRAPH_FILENAMES = (
    '__main__.py',
    'main.py',
)
DEFAULT_GRAPH_ATTR = 'get_graph'


def get_default_services(filename, services=None):
    dirname = os.path.dirname(filename)
    services_filename = os.path.join(dirname, DEFAULT_SERVICES_FILENAME)
    if os.path.exists(services_filename):
        with open(services_filename) as file:
            code = compile(file.read(), services_filename, 'exec')
        context = {
            '__name__': '__bonobo__',
            '__file__': services_filename,
        }
        exec(code, context)

        return {
            **context[DEFAULT_SERVICES_ATTR](),
            **(services or {}),
        }
    return services or {}


def _install_requirements(requirements):
    """Install requirements given a path to requirements.txt file."""
    import importlib
    import pip

    pip.main(['install', '-r', requirements])
    # Some shenanigans to be sure everything is importable after this, especially .egg-link files which
    # are referenced in *.pth files and apparently loaded by site.py at some magic bootstrap moment of the
    # python interpreter.
    pip.utils.pkg_resources = importlib.reload(pip.utils.pkg_resources)
    import site
    importlib.reload(site)


def read(filename, module, install=False, quiet=False, verbose=False, env=None):
    import re
    import runpy
    from bonobo import Graph, settings

    if quiet:
        settings.QUIET.set(True)

    if verbose:
        settings.DEBUG.set(True)

    if env:
        quote_killer = re.compile('["\']')
        for e in env:
            var_name, var_value = e.split('=')
            os.environ[var_name] = quote_killer.sub('', var_value)

    if filename:
        if os.path.isdir(filename):
            if install:
                requirements = os.path.join(filename, 'requirements.txt')
                _install_requirements(requirements)

            pathname = filename
            for filename in DEFAULT_GRAPH_FILENAMES:
                filename = os.path.join(pathname, filename)
                if os.path.exists(filename):
                    break
            if not os.path.exists(filename):
                raise IOError('Could not find entrypoint (candidates: {}).'.format(', '.join(DEFAULT_GRAPH_FILENAMES)))
        elif install:
            requirements = os.path.join(os.path.dirname(filename), 'requirements.txt')
            _install_requirements(requirements)
        context = runpy.run_path(filename, run_name='__bonobo__')
    elif module:
        context = runpy.run_module(module, run_name='__bonobo__')
        filename = context['__file__']
    else:
        raise RuntimeError('UNEXPECTED: argparse should not allow this.')

    graphs = dict((k, v) for k, v in context.items() if isinstance(v, Graph))

    assert len(graphs) == 1, (
        'Having zero or more than one graph definition in one file is unsupported for now, '
        'but it is something that will be implemented in the future.\n\nExpected: 1, got: {}.'
    ).format(len(graphs))

    graph = list(graphs.values())[0]
    plugins = []
    services = get_default_services(
        filename, context.get(DEFAULT_SERVICES_ATTR)() if DEFAULT_SERVICES_ATTR in context else None
    )

    return graph, plugins, services


def execute(filename, module, install=False, quiet=False, verbose=False, env=None):
    graph, plugins, services = read(filename, module, install, quiet, verbose, env)

    return bonobo.run(graph, plugins=plugins, services=services)


def register_generic_run_arguments(parser, required=True):
    source_group = parser.add_mutually_exclusive_group(required=required)
    source_group.add_argument('filename', nargs='?', type=str)
    source_group.add_argument('--module', '-m', type=str)
    parser.add_argument('--env', '-e', action='append')
    return parser


def register(parser):
    parser = register_generic_run_arguments(parser)
    verbosity_group = parser.add_mutually_exclusive_group()
    verbosity_group.add_argument('--quiet', '-q', action='store_true')
    verbosity_group.add_argument('--verbose', '-v', action='store_true')
    parser.add_argument('--install', '-I', action='store_true')
    return execute



import inspect
import os
import runpy


class _RequiredModule:
    def __init__(self, dct):
        self.__dict__ = dct


class _RequiredModulesRegistry(dict):
    def require(self, name):
        if name not in self:
            bits = name.split('.')
            pathname = os.path.join(os.getcwd(), os.path.dirname(inspect.getfile(inspect.stack()[1][0])))
            filename = os.path.join(pathname, *bits[:-1], bits[-1] + '.py')
            self[name] = _RequiredModule(runpy.run_path(filename, run_name=name))
        return self[name]


registry = _RequiredModulesRegistry()
require = registry.require



from contextlib import contextmanager
from unittest.mock import MagicMock

from bonobo import open_fs
from bonobo.execution.node import NodeExecutionContext


class CapturingNodeExecutionContext(NodeExecutionContext):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.send = MagicMock()


@contextmanager
def optional_contextmanager(cm, *, ignore=False):
    if cm is None or ignore:
        yield
    else:
        with cm:
            yield


class FilesystemTester:
    def __init__(self, extension='txt', mode='w'):
        self.extension = extension
        self.input_data = ''
        self.mode = mode

    def get_services_for_reader(self, tmpdir):
        fs, filename = open_fs(tmpdir), 'input.' + self.extension
        with fs.open(filename, self.mode) as fp:
            fp.write(self.input_data)
        return fs, filename, {'fs': fs}

    def get_services_for_writer(self, tmpdir):
        fs, filename = open_fs(tmpdir), 'output.' + self.extension
        return fs, filename, {'fs': fs}



import mimetypes
import os

import bonobo

SHORTCUTS = {
    'csv': 'text/csv',
    'json': 'application/json',
    'pickle': 'pickle',
    'plain': 'text/plain',
    'text': 'text/plain',
    'txt': 'text/plain',
}

REGISTRY = {
    'application/json': (bonobo.JsonReader, bonobo.JsonWriter),
    'pickle': (bonobo.PickleReader, bonobo.PickleWriter),
    'text/csv': (bonobo.CsvReader, bonobo.CsvWriter),
    'text/plain': (bonobo.FileReader, bonobo.FileWriter),
}

READER = 'reader'
WRITER = 'writer'


def resolve_factory(name, filename, factory_type):
    """
    Try to resolve which transformation factory to use for this filename. User eventually provided a name, which has
    priority, otherwise we try to detect it using the mimetype detection on filename.

    """
    if name is None:
        name = mimetypes.guess_type(filename)[0]

    if name in SHORTCUTS:
        name = SHORTCUTS[name]

    if name is None:
        _, _ext = os.path.splitext(filename)
        if _ext:
            _ext = _ext[1:]
        if _ext in SHORTCUTS:
            name = SHORTCUTS[_ext]

    if not name in REGISTRY:
        raise RuntimeError(
            'Could not resolve {factory_type} factory for {filename} ({name}). Try providing it explicitely using -{opt} <format>.'.
            format(name=name, filename=filename, factory_type=factory_type, opt=factory_type[0])
        )

    if factory_type == READER:
        return REGISTRY[name][0]
    elif factory_type == WRITER:
        return REGISTRY[name][1]
    else:
        raise ValueError('Invalid factory type.')


def execute(input, output, reader=None, reader_options=None, writer=None, writer_options=None, options=None):
    reader = resolve_factory(reader, input, READER)(input)
    writer = resolve_factory(writer, output, WRITER)(output)

    graph = bonobo.Graph()
    graph.add_chain(reader, writer)

    return bonobo.run(
        graph, services={
            'fs': bonobo.open_fs(),
        }
    )


def register(parser):
    parser.add_argument('input')
    parser.add_argument('output')
    parser.add_argument('--' + READER, '-r')
    parser.add_argument('--' + WRITER, '-w')
    # parser.add_argument('--reader-option', '-ro', dest='reader_options')
    # parser.add_argument('--writer-option', '-wo', dest='writer_options')
    # parser.add_argument('--option', '-o', dest='options')
    return execute



import traceback
from queue import Empty
from time import sleep

from bonobo.constants import INHERIT_INPUT, NOT_MODIFIED
from bonobo.errors import InactiveReadableError, UnrecoverableError
from bonobo.execution.base import LoopingExecutionContext
from bonobo.structs.bags import Bag
from bonobo.structs.inputs import Input
from bonobo.util.compat import deprecated_alias
from bonobo.util.inspect import iserrorbag, isloopbackbag
from bonobo.util.iterators import iter_if_not_sequence
from bonobo.util.objects import get_name
from bonobo.util.statistics import WithStatistics


class NodeExecutionContext(WithStatistics, LoopingExecutionContext):
    """
    todo: make the counter dependant of parent context?
    """

    @property
    def alive(self):
        """todo check if this is right, and where it is used"""
        return self._started and not self._stopped

    @property
    def alive_str(self):
        return '+' if self.alive else '-'

    def __init__(self, wrapped, parent=None, services=None):
        LoopingExecutionContext.__init__(self, wrapped, parent=parent, services=services)
        WithStatistics.__init__(self, 'in', 'out', 'err')

        self.input = Input()
        self.outputs = []

    def __str__(self):
        return self.alive_str + ' ' + self.__name__ + self.get_statistics_as_string(prefix=' ')

    def __repr__(self):
        name, type_name = get_name(self), get_name(type(self))
        return '<{}({}{}){}>'.format(type_name, self.alive_str, name, self.get_statistics_as_string(prefix=' '))

    def write(self, *messages):
        """
        Push a message list to this context's input queue.

        :param mixed value: message
        """
        for message in messages:
            self.input.put(message)

    # XXX deprecated alias
    recv = deprecated_alias('recv', write)

    def send(self, value, _control=False):
        """
        Sends a message to all of this context's outputs.

        :param mixed value: message
        :param _control: if true, won't count in statistics.
        """

        if not _control:
            self.increment('out')

        if iserrorbag(value):
            value.apply(self.handle_error)
        elif isloopbackbag(value):
            self.input.put(value)
        else:
            for output in self.outputs:
                output.put(value)

    push = deprecated_alias('push', send)

    def get(self):  # recv() ? input_data = self.receive()
        """
        Get from the queue first, then increment stats, so if Queue raise Timeout or Empty, stat won't be changed.

        """
        row = self.input.get(timeout=self.PERIOD)
        self.increment('in')
        return row

    def loop(self):
        while True:
            try:
                self.step()
            except KeyboardInterrupt:
                raise
            except InactiveReadableError:
                break
            except Empty:
                sleep(self.PERIOD)
                continue
            except UnrecoverableError as exc:
                self.handle_error(exc, traceback.format_exc())
                self.input.shutdown()
                break
            except Exception as exc:  # pylint: disable=broad-except
                self.handle_error(exc, traceback.format_exc())

    def step(self):
        # Pull data from the first available input channel.
        """Runs a transformation callable with given args/kwargs and flush the result into the right
        output channel."""

        input_bag = self.get()

        # todo add timer
        self.handle_results(input_bag, input_bag.apply(self._stack))

    def handle_results(self, input_bag, results):
        # self._exec_time += timer.duration
        # Put data onto output channels
        try:
            results = iter_if_not_sequence(results)
        except TypeError:  # not an iterator
            if results:
                self.send(_resolve(input_bag, results))
            else:
                # case with no result, an execution went through anyway, use for stats.
                # self._exec_count += 1
                pass
        else:
            while True:  # iterator
                try:
                    result = next(results)
                except StopIteration:
                    break
                else:
                    self.send(_resolve(input_bag, result))


def _resolve(input_bag, output):
    # NotModified means to send the input unmodified to output.
    if output is NOT_MODIFIED:
        return input_bag

    if iserrorbag(output):
        return output

    # If it does not look like a bag, let's create one for easier manipulation
    if hasattr(output, 'apply'):
        # Already a bag? Check if we need to set parent.
        if INHERIT_INPUT in output.flags:
            output.set_parent(input_bag)
    else:
        # Not a bag? Let's encapsulate it.
        output = Bag(output)

    return output



from functools import partial

from bonobo.config import create_container
from bonobo.constants import BEGIN, END
from bonobo.execution.node import NodeExecutionContext
from bonobo.execution.plugin import PluginExecutionContext


class GraphExecutionContext:
    @property
    def started(self):
        return any(node.started for node in self.nodes)

    @property
    def stopped(self):
        return all(node.started and node.stopped for node in self.nodes)

    @property
    def alive(self):
        return any(node.alive for node in self.nodes)

    def __init__(self, graph, plugins=None, services=None):
        self.graph = graph
        self.nodes = [NodeExecutionContext(node, parent=self) for node in self.graph]
        self.plugins = [PluginExecutionContext(plugin, parent=self) for plugin in plugins or ()]
        self.services = create_container(services)

        # Probably not a good idea to use it unless you really know what you're doing. But you can access the context.
        self.services['__graph_context'] = self

        for i, node_context in enumerate(self):
            node_context.outputs = [self[j].input for j in self.graph.outputs_of(i)]
            node_context.input.on_begin = partial(node_context.send, BEGIN, _control=True)
            node_context.input.on_end = partial(node_context.send, END, _control=True)
            node_context.input.on_finalize = partial(node_context.stop)

    def __getitem__(self, item):
        return self.nodes[item]

    def __len__(self):
        return len(self.nodes)

    def __iter__(self):
        yield from self.nodes

    def write(self, *messages):
        """Push a list of messages in the inputs of this graph's inputs, matching the output of special node "BEGIN" in
        our graph."""

        for i in self.graph.outputs_of(BEGIN):
            for message in messages:
                self[i].write(message)

    def start(self):
        # todo use strategy
        for node in self.nodes:
            node.start()

    def stop(self):
        # todo use strategy
        for node in self.nodes:
            node.stop()

    def loop(self):
        # todo use strategy
        for node in self.nodes:
            node.loop()



import logging
from os import environ
from unittest.mock import patch

import pytest

from bonobo import settings

TEST_SETTING = 'TEST_SETTING'


def test_to_bool():
    assert not settings.to_bool('')
    assert not settings.to_bool('FALSE')
    assert not settings.to_bool('NO')
    assert not settings.to_bool('0')

    assert settings.to_bool('yup')
    assert settings.to_bool('True')
    assert settings.to_bool('yes')
    assert settings.to_bool('1')


def test_setting():
    s = settings.Setting(TEST_SETTING)
    assert s.get() is None

    with patch.dict(environ, {TEST_SETTING: 'hello'}):
        assert s.get() is None
        s.clear()
        assert s.get() == 'hello'

    s = settings.Setting(TEST_SETTING, default='nope')
    assert s.get() is 'nope'

    with patch.dict(environ, {TEST_SETTING: 'hello'}):
        assert s.get() == 'nope'
        s.clear()
        assert s.get() == 'hello'


def test_default_settings():
    settings.clear_all()

    assert settings.DEBUG.get() == False
    assert settings.PROFILE.get() == False
    assert settings.QUIET.get() == False
    assert settings.LOGGING_LEVEL.get() == logging._checkLevel('INFO')

    with patch.dict(environ, {'DEBUG': 't'}):
        settings.clear_all()
        assert settings.LOGGING_LEVEL.get() == logging._checkLevel('DEBUG')

    settings.clear_all()


def test_check():
    settings.check()
    with patch.dict(environ, {'DEBUG': 't', 'PROFILE': 't', 'QUIET': 't'}):
        settings.clear_all()
        with pytest.raises(RuntimeError):
            settings.check()
    settings.clear_all()



import bonobo
import time

from bonobo.constants import NOT_MODIFIED


def pause(*args, **kwargs):
    time.sleep(0.1)
    return NOT_MODIFIED


graph = bonobo.Graph(
    lambda: tuple(range(20)),
    pause,
    print,
)

if __name__ == '__main__':
    bonobo.run(graph)



"""
Example on how to use :class:`bonobo.Bag` instances to pass flexible args/kwargs to the next callable.

.. graphviz::

    digraph {
        rankdir = LR;
        stylesheet = "../_static/graphs.css";

        BEGIN [shape="point"];
        BEGIN -> "extract()" -> "transform(...)" -> "load(...)";
    }

"""

from random import randint

from bonobo import Bag, Graph


def extract():
    yield Bag(topic='foo')
    yield Bag(topic='bar')
    yield Bag(topic='baz')


def transform(topic: str):
    return Bag.inherit(title=topic.title(), rand=randint(10, 99))


def load(topic: str, title: str, rand: int):
    print('{} ({}) wait={}'.format(title, topic, rand))


graph = Graph()
graph.add_chain(extract, transform, load)

if __name__ == '__main__':
    from bonobo import run

    run(graph)



import time
import traceback
from concurrent.futures import Executor, ProcessPoolExecutor, ThreadPoolExecutor

from bonobo.constants import BEGIN, END
from bonobo.strategies.base import Strategy
from bonobo.structs.bags import Bag
from bonobo.util.errors import print_error


class ExecutorStrategy(Strategy):
    """
    Strategy based on a concurrent.futures.Executor subclass (or similar interface).

    """

    executor_factory = Executor

    def create_executor(self):
        return self.executor_factory()

    def execute(self, graph, *args, plugins=None, services=None, **kwargs):
        context = self.create_graph_execution_context(graph, plugins=plugins, services=services)
        context.write(BEGIN, Bag(), END)

        executor = self.create_executor()

        futures = []

        for plugin_context in context.plugins:

            def _runner(plugin_context=plugin_context):
                with plugin_context:
                    try:
                        plugin_context.loop()
                    except Exception as exc:
                        print_error(exc, traceback.format_exc(), context=plugin_context)

            futures.append(executor.submit(_runner))

        for node_context in context.nodes:

            def _runner(node_context=node_context):
                try:
                    node_context.start()
                except Exception as exc:
                    print_error(exc, traceback.format_exc(), context=node_context, method='start')
                    node_context.input.on_end()
                else:
                    node_context.loop()

                try:
                    node_context.stop()
                except Exception as exc:
                    print_error(exc, traceback.format_exc(), context=node_context, method='stop')

            futures.append(executor.submit(_runner))

        while context.alive:
            time.sleep(0.1)

        for plugin_context in context.plugins:
            plugin_context.shutdown()

        executor.shutdown()

        return context


class ThreadPoolExecutorStrategy(ExecutorStrategy):
    executor_factory = ThreadPoolExecutor


class ProcessPoolExecutorStrategy(ExecutorStrategy):
    executor_factory = ProcessPoolExecutor



from bonobo.constants import BEGIN, END
from bonobo.strategies.base import Strategy
from bonobo.structs.bags import Bag


class NaiveStrategy(Strategy):
    def execute(self, graph, *args, plugins=None, **kwargs):
        context = self.create_graph_execution_context(graph, plugins=plugins)
        context.write(BEGIN, Bag(), END)

        # TODO: how to run plugins in "naive" mode ?
        context.start()
        context.loop()
        context.stop()

        return context



from bonobo.execution.graph import GraphExecutionContext


class Strategy:
    """
    Base class for execution strategies.

    """
    graph_execution_context_factory = GraphExecutionContext

    def create_graph_execution_context(self, graph, *args, **kwargs):
        return self.graph_execution_context_factory(graph, *args, **kwargs)

    def execute(self, graph, *args, **kwargs):
        raise NotImplementedError



def execute(name, branch):
    try:
        from cookiecutter.main import cookiecutter
    except ImportError as exc:
        raise ImportError(
            'You must install "cookiecutter" to use this command.\n\n $ pip install cookiecutter\n'
        ) from exc

    return cookiecutter(
        'https://github.com/python-bonobo/cookiecutter-bonobo.git',
        extra_context={'name': name},
        no_input=True,
        checkout=branch
    )


def register(parser):
    parser.add_argument('name')
    parser.add_argument('--branch', '-b', default='master')
    return execute



"""
Extracts a list of parisian bars where you can buy a coffee for a reasonable price, and store them in a flat text file.

.. graphviz::

    digraph {
        rankdir = LR;
        stylesheet = "../_static/graphs.css";

        BEGIN [shape="point"];
        BEGIN -> "ODS()" -> "transform" -> "FileWriter()";
    }

"""

import bonobo
from bonobo.commands.run import get_default_services
from bonobo.ext.opendatasoft import OpenDataSoftAPI

filename = 'coffeeshops.txt'

graph = bonobo.Graph(
    OpenDataSoftAPI(dataset='liste-des-cafes-a-un-euro', netloc='opendata.paris.fr'),
    lambda row: '{nom_du_cafe}, {adresse}, {arrondissement} Paris, France'.format(**row),
    bonobo.FileWriter(path=filename),
)

if __name__ == '__main__':
    bonobo.run(graph, services=get_default_services(__file__))



from bonobo.config import Option
from bonobo.config.processors import ContextProcessor
from bonobo.constants import NOT_MODIFIED
from bonobo.nodes.io.base import FileHandler, Reader, Writer
from bonobo.util.objects import ValueHolder


class FileReader(Reader, FileHandler):
    """Component factory for file-like readers.

    On its own, it can be used to read a file and yield one row per line, trimming the "eol" character at the end if
    present. Extending it is usually the right way to create more specific file readers (like json, csv, etc.)
    """

    mode = Option(str, default='r')

    def read(self, fs, file):
        """
        Write a row on the next line of given file.
        Prefix is used for newlines.
        """
        for line in file:
            yield line.rstrip(self.eol)


class FileWriter(Writer, FileHandler):
    """Component factory for file or file-like writers.

    On its own, it can be used to write in a file one line per row that comes into this component. Extending it is
    usually the right way to create more specific file writers (like json, csv, etc.)
    """

    mode = Option(str, default='w+')

    @ContextProcessor
    def lineno(self, context, fs, file):
        lineno = ValueHolder(0)
        yield lineno

    def write(self, fs, file, lineno, line):
        """
        Write a row on the next line of opened file in context.
        """
        self._write_line(file, (self.eol if lineno.value else '') + line)
        lineno += 1
        return NOT_MODIFIED

    def _write_line(self, file, line):
        return file.write(line)



import pickle

from bonobo.config import Option
from bonobo.config.processors import ContextProcessor
from bonobo.constants import NOT_MODIFIED
from bonobo.nodes.io.base import FileHandler, IOFormatEnabled
from bonobo.nodes.io.file import FileReader, FileWriter
from bonobo.util.objects import ValueHolder


class PickleHandler(FileHandler):
    """

    .. attribute:: item_names

        The names of the items in the pickle, if it is not defined in the first item of the pickle.

    """

    item_names = Option(tuple, required=False)


class PickleReader(IOFormatEnabled, FileReader, PickleHandler):
    """
    Reads a Python pickle object and yields the items in dicts.
    """

    mode = Option(str, default='rb')

    @ContextProcessor
    def pickle_headers(self, context, fs, file):
        yield ValueHolder(self.item_names)

    def read(self, fs, file, pickle_headers):
        data = pickle.load(file)

        # if the data is not iterable, then wrap the object in a list so it may be iterated
        if isinstance(data, dict):
            is_dict = True
            iterator = iter(data.items())
        else:
            is_dict = False
            try:
                iterator = iter(data)
            except TypeError:
                iterator = iter([data])

        if not pickle_headers.get():
            pickle_headers.set(next(iterator))

        item_count = len(pickle_headers.value)

        for i in iterator:
            if len(i) != item_count:
                raise ValueError('Received an object with %d items, expecting %d.' % (
                    len(i),
                    item_count,
                ))

            yield self.get_output(dict(zip(i)) if is_dict else dict(zip(pickle_headers.value, i)))


class PickleWriter(IOFormatEnabled, FileWriter, PickleHandler):
    mode = Option(str, default='wb')

    def write(self, fs, file, lineno, item):
        """
        Write a pickled item to the opened file.
        """
        file.write(pickle.dumps(item))
        lineno += 1
        return NOT_MODIFIED



import bonobo
from bonobo.commands.run import get_default_services


def skip_comments(line):
    if not line.startswith('#'):
        yield line


graph = bonobo.Graph(
    bonobo.FileReader('datasets/passwd.txt'),
    skip_comments,
    lambda s: s.split(':'),
    lambda l: l[0],
    print,
)

if __name__ == '__main__':
    bonobo.run(graph, services=get_default_services(__file__))



from bonobo.config import Configurable
from bonobo.util.objects import get_attribute_or_create


class Plugin:
    """
    A plugin is an extension to the core behavior of bonobo. If you're writing transformations, you should not need
    to use this interface.
    
    For examples, you can read bonobo.ext.console.ConsoleOutputPlugin, or bonobo.ext.jupyter.JupyterOutputPlugin that
    respectively permits an interactive output on an ANSI console and a rich output in a jupyter notebook.
    
    Warning: THE PLUGIN API IS PRE-ALPHA AND WILL EVOLVE BEFORE 1.0, DO NOT RELY ON IT BEING STABLE!
    
    """

    def __init__(self, context):
        self.context = context

    def initialize(self):
        pass

    def run(self):
        pass

    def finalize(self):
        pass


def get_enhancers(obj):
    try:
        return get_attribute_or_create(obj, '__enhancers__', list())
    except AttributeError:
        return list()


class NodeEnhancer(Configurable):
    def __matmul__(self, other):
        get_enhancers(other).append(self)
        return other



import bonobo
from bonobo.commands.run import get_default_services

graph = bonobo.Graph(
    bonobo.CsvReader('datasets/coffeeshops.txt', headers=('item', )),
    bonobo.PrettyPrinter(),
)

if __name__ == '__main__':
    bonobo.run(graph, services=get_default_services(__file__))



# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class AbstractError(NotImplementedError):
    """Abstract error is a convenient error to declare a method as "being left as an exercise for the reader"."""

    def __init__(self, method):
        super().__init__(
            'Call to abstract method {class_name}.{method_name}(...): missing implementation.'.format(
                class_name=method.__self__.__name__,
                method_name=method.__name__,
            )
        )


class InactiveIOError(IOError):
    pass


class InactiveReadableError(InactiveIOError):
    pass


class InactiveWritableError(InactiveIOError):
    pass


class ValidationError(RuntimeError):
    def __init__(self, inst, message):
        super(ValidationError, self).__init__(
            'Validation error in {class_name}: {message}'.format(
                class_name=type(inst).__name__,
                message=message,
            )
        )


class ProhibitedOperationError(RuntimeError):
    pass


class ConfigurationError(Exception):
    pass


class UnrecoverableError(Exception):
    """Flag for errors that must interrupt the workflow, either because they will happen for sure on each node run, or
    because you know that your transformation has no point continuing runnning after a bad event."""


class UnrecoverableValueError(UnrecoverableError, ValueError):
    pass


class UnrecoverableRuntimeError(UnrecoverableError, RuntimeError):
    pass


class UnrecoverableNotImplementedError(UnrecoverableError, NotImplementedError):
    pass


class MissingServiceImplementationError(UnrecoverableError, KeyError):
    pass



import json

from bonobo.config.processors import ContextProcessor
from bonobo.constants import NOT_MODIFIED
from bonobo.nodes.io.base import FileHandler, IOFormatEnabled
from bonobo.nodes.io.file import FileReader, FileWriter
from bonobo.structs.bags import Bag


class JsonHandler(FileHandler):
    eol = ',\n'
    prefix, suffix = '[', ']'


class JsonReader(IOFormatEnabled, FileReader, JsonHandler):
    loader = staticmethod(json.load)

    def read(self, fs, file):
        for line in self.loader(file):
            yield self.get_output(line)


class JsonDictItemsReader(JsonReader):
    def read(self, fs, file):
        for line in self.loader(file).items():
            yield Bag(*line)


class JsonWriter(IOFormatEnabled, FileWriter, JsonHandler):
    @ContextProcessor
    def envelope(self, context, fs, file, lineno):
        file.write(self.prefix)
        yield
        file.write(self.suffix)

    def write(self, fs, file, lineno, *args, **kwargs):
        """
        Write a json row on the next line of file pointed by ctx.file.

        :param ctx:
        :param row:
        """
        row = self.get_input(*args, **kwargs)
        self._write_line(file, (self.eol if lineno.value else '') + json.dumps(row))
        lineno += 1
        return NOT_MODIFIED



import functools
import warnings
from functools import partial

from bonobo import Bag
from bonobo.config import Configurable, Method

_isarg = lambda item: type(item) is int
_iskwarg = lambda item: type(item) is str


class Operation():
    def __init__(self, item, callable):
        self.item = item
        self.callable = callable

    def __repr__(self):
        return '<operation {} on {}>'.format(self.callable.__name__, self.item)

    def apply(self, *args, **kwargs):
        if _isarg(self.item):
            return (*args[0:self.item], self.callable(args[self.item]), *args[self.item + 1:]), kwargs
        if _iskwarg(self.item):
            return args, {**kwargs, self.item: self.callable(kwargs.get(self.item))}
        raise RuntimeError('Houston, we have a problem...')


class FactoryOperation():
    def __init__(self, factory, callable):
        self.factory = factory
        self.callable = callable

    def __repr__(self):
        return '<factory operation {}>'.format(self.callable.__name__)

    def apply(self, *args, **kwargs):
        return self.callable(*args, **kwargs)


CURSOR_TYPES = {}


def operation(mixed):
    def decorator(m, ctype=mixed):
        def lazy_operation(self, *args, **kwargs):
            @functools.wraps(m)
            def actual_operation(x):
                return m(self, x, *args, **kwargs)

            self.factory.operations.append(Operation(self.item, actual_operation))
            return CURSOR_TYPES[ctype](self.factory, self.item) if ctype else self

        return lazy_operation

    return decorator if isinstance(mixed, str) else decorator(mixed, ctype=None)


def factory_operation(m):
    def lazy_operation(self, *config):
        @functools.wraps(m)
        def actual_operation(*args, **kwargs):
            return m(self, *config, *args, **kwargs)

        self.operations.append(FactoryOperation(self, actual_operation))
        return self

    return lazy_operation


class Cursor():
    _type = None

    def __init__(self, factory, item):
        self.factory = factory
        self.item = item

    @operation('dict')
    def dict(self, x):
        return x if isinstance(x, dict) else dict(x)

    @operation('int')
    def int(self):
        pass

    @operation('str')
    def str(self, x):
        return x if isinstance(x, str) else str(x)

    @operation('list')
    def list(self):
        pass

    @operation('tuple')
    def tuple(self):
        pass

    def __getattr__(self, item):
        """
        Fallback to type methods if they exist, for example StrCursor.upper will use str.upper if not overriden, etc.

        :param item:
        """
        if self._type and item in self._type.__dict__:
            method = self._type.__dict__[item]

            @operation
            @functools.wraps(method)
            def _operation(self, x, *args, **kwargs):
                return method(x, *args, **kwargs)

            setattr(self, item, partial(_operation, self))
            return getattr(self, item)

        raise AttributeError('Unknown operation {}.{}().'.format(
            type(self).__name__,
            item,
        ))


CURSOR_TYPES['default'] = Cursor


class DictCursor(Cursor):
    _type = dict

    @operation('default')
    def get(self, x, path):
        return x.get(path)

    @operation
    def map_keys(self, x, mapping):
        return {mapping.get(k): v for k, v in x.items()}


CURSOR_TYPES['dict'] = DictCursor


class StringCursor(Cursor):
    _type = str


CURSOR_TYPES['str'] = StringCursor


class Factory(Configurable):
    initialize = Method(required=False)

    def __init__(self, *args, **kwargs):
        warnings.warn(
            __file__ +
            ' is experimental, API may change in the future, use it as a preview only and knowing the risks.',
            FutureWarning
        )
        super(Factory, self).__init__(*args, **kwargs)
        self.default_cursor_type = 'default'
        self.operations = []

        if self.initialize is not None:
            self.initialize(self)

    @factory_operation
    def move(self, _from, _to, *args, **kwargs):
        if _from == _to:
            return args, kwargs

        if _isarg(_from):
            value = args[_from]
            args = args[:_from] + args[_from + 1:]
        elif _iskwarg(_from):
            value = kwargs[_from]
            kwargs = {k: v for k, v in kwargs if k != _from}
        else:
            raise RuntimeError('Houston, we have a problem...')

        if _isarg(_to):
            return (*args[:_to], value, *args[_to + 1:]), kwargs
        elif _iskwarg(_to):
            return args, {**kwargs, _to: value}
        else:
            raise RuntimeError('Houston, we have a problem...')

    def __call__(self, *args, **kwargs):
        print('factory call on', args, kwargs)
        for operation in self.operations:
            args, kwargs = operation.apply(*args, **kwargs)
            print(' ... after', operation, 'got', args, kwargs)
        return Bag(*args, **kwargs)

    def __getitem__(self, item):
        return CURSOR_TYPES[self.default_cursor_type](self, item)


if __name__ == '__main__':
    f = Factory()

    f[0].dict().map_keys({'foo': 'F00'})
    f['foo'].str().upper()

    print('operations:', f.operations)
    print(f({'foo': 'bisou'}, foo='blah'))
'''
specs:

- rename keys of an input dict (in args, or kwargs) using a translation map.


f = Factory()

f[0]
f['xxx'] = 
    
f[0].dict().get('foo.bar').move_to('foo.baz').apply(str.upper)
f[0].get('foo.*').items().map(str.lower)

f['foo'].keys_map({
    'a': 'b'
})

'''



import bonobo

from bonobo import Filter


class OddOnlyFilter(Filter):
    def filter(self, i):
        return i % 2


@Filter
def multiples_of_three(i):
    return not (i % 3)


graph = bonobo.Graph(
    lambda: tuple(range(50)),
    OddOnlyFilter(),
    multiples_of_three,
    print,
)

if __name__ == '__main__':
    bonobo.run(graph)



import logging
import sys
import textwrap
from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING

from colorama import Fore, Style

from bonobo import settings
from bonobo.util.term import CLEAR_EOL

iswindows = (sys.platform == 'win32')


def get_format():
    yield '{b}[%(fg)s%(levelname)s{b}][{w}'
    yield '{b}][{w}'.join(('%(spent)04d', '%(name)s'))
    yield '{b}]'
    yield ' %(fg)s%(message)s{r}'
    if not iswindows:
        yield CLEAR_EOL


colors = {
    'b': '' if iswindows else Fore.BLACK,
    'w': '' if iswindows else Fore.LIGHTBLACK_EX,
    'r': '' if iswindows else Style.RESET_ALL,
}
format = (''.join(get_format())).format(**colors)


class Filter(logging.Filter):
    def filter(self, record):
        record.spent = record.relativeCreated // 1000
        if iswindows:
            record.fg = ''
        elif record.levelname == 'DEBG':
            record.fg = Fore.LIGHTBLACK_EX
        elif record.levelname == 'INFO':
            record.fg = Fore.LIGHTWHITE_EX
        elif record.levelname == 'WARN':
            record.fg = Fore.LIGHTYELLOW_EX
        elif record.levelname == 'ERR ':
            record.fg = Fore.LIGHTRED_EX
        elif record.levelname == 'CRIT':
            record.fg = Fore.RED
        else:
            record.fg = Fore.LIGHTWHITE_EX
        return True


class Formatter(logging.Formatter):
    def formatException(self, ei):
        tb = super().formatException(ei)
        if iswindows:
            return textwrap.indent(tb, ' | ')
        else:
            return textwrap.indent(tb, Fore.BLACK + ' | ' + Fore.WHITE)


def setup(level):
    logging.addLevelName(DEBUG, 'DEBG')
    logging.addLevelName(INFO, 'INFO')
    logging.addLevelName(WARNING, 'WARN')
    logging.addLevelName(ERROR, 'ERR ')
    logging.addLevelName(CRITICAL, 'CRIT')
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(Formatter(format))
    handler.addFilter(Filter())
    root = logging.getLogger()
    root.addHandler(handler)
    root.setLevel(level)


def set_level(level):
    logging.getLogger().setLevel(level)


def get_logger(name='bonobo'):
    return logging.getLogger(name)


# Compatibility with python logging
getLogger = get_logger

# Setup formating and level.
setup(level=settings.LOGGING_LEVEL.get())



import inspect


def test_wildcard_import():
    bonobo = __import__('bonobo')
    assert bonobo.__version__

    for name in dir(bonobo):
        # ignore attributes starting by underscores
        if name.startswith('_'):
            continue
        attr = getattr(bonobo, name)
        if inspect.ismodule(attr):
            continue

        assert name in bonobo.__all__



import argparse

from bonobo import logging, settings

logger = logging.get_logger()


def entrypoint(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', '-D', action='store_true')

    subparsers = parser.add_subparsers(dest='command')
    subparsers.required = True

    commands = {}

    def register_extension(ext, commands=commands):
        try:
            parser = subparsers.add_parser(ext.name)
            commands[ext.name] = ext.plugin(parser)
        except Exception:
            logger.exception('Error while loading command {}.'.format(ext.name))

    from stevedore import ExtensionManager
    mgr = ExtensionManager(namespace='bonobo.commands')
    mgr.map(register_extension)

    args = parser.parse_args(args).__dict__
    if args.pop('debug', False):
        settings.DEBUG.set(True)
        settings.LOGGING_LEVEL.set(logging.DEBUG)
        logging.set_level(settings.LOGGING_LEVEL.get())

    logger.debug('Command: ' + args['command'] + ' Arguments: ' + repr(args))
    commands[args.pop('command')](**args)



from bonobo.structs.tokens import Token

BEGIN = Token('Begin')
END = Token('End')
INHERIT_INPUT = Token('InheritInput')
LOOPBACK = Token('Loopback')
NOT_MODIFIED = Token('NotModified')
DEFAULT_SERVICES_FILENAME = '_services.py'
DEFAULT_SERVICES_ATTR = 'get_services'



import bonobo
from bonobo import Bag
from bonobo.commands.run import get_default_services


def get_fields(**row):
    return Bag(**row['fields'])


graph = bonobo.Graph(
    bonobo.JsonReader('datasets/theaters.json'),
    get_fields,
    bonobo.PrettyPrinter(),
)

if __name__ == '__main__':
    bonobo.run(graph, services=get_default_services(__file__))



from bonobo.config.configurables import Configurable
from bonobo.config.options import Method, Option
from bonobo.config.processors import ContextProcessor
from bonobo.config.services import Container, Exclusive, Service, requires, create_container

use = requires

# Bonobo's Config API
__all__ = [
    'Configurable',
    'Container',
    'ContextProcessor',
    'Exclusive',
    'Method',
    'Option',
    'Service',
    'create_container',
    'requires',
    'use',
]



import itertools

from bonobo.constants import INHERIT_INPUT, LOOPBACK

__all__ = [
    'Bag',
    'ErrorBag',
]


class Bag:
    """
    Bags are simple datastructures that holds arguments and keyword arguments together, that may be applied to a
    callable.
    
    Example:
    
        >>> from bonobo import Bag
        >>> def myfunc(foo, *, bar):
        ...     print(foo, bar)
        ...
        >>> bag = Bag('foo', bar='baz')
        >>> bag.apply(myfunc)
        foo baz
    
    A bag can inherit another bag, allowing to override only a few arguments without touching the parent.
    
    Example:
        
        >>> bag2 = Bag(bar='notbaz', _parent=bag)
        >>> bag2.apply(myfunc)
        foo notbaz
    
    """

    default_flags = ()

    def __init__(self, *args, _flags=None, _parent=None, **kwargs):
        self._flags = type(self).default_flags + (_flags or ())
        self._parent = _parent
        self._args = args
        self._kwargs = kwargs

    @property
    def args(self):
        if self._parent is None:
            return self._args
        return (
            *self._parent.args,
            *self._args,
        )

    @property
    def kwargs(self):
        if self._parent is None:
            return self._kwargs
        return {
            **self._parent.kwargs,
            **self._kwargs,
        }

    @property
    def flags(self):
        return self._flags

    def apply(self, func_or_iter, *args, **kwargs):
        if callable(func_or_iter):
            return func_or_iter(*args, *self.args, **kwargs, **self.kwargs)

        if len(args) == 0 and len(kwargs) == 0:
            try:
                iter(func_or_iter)

                def generator():
                    yield from func_or_iter

                return generator()
            except TypeError as exc:
                raise TypeError('Could not apply bag to {}.'.format(func_or_iter)) from exc

        raise TypeError('Could not apply bag to {}.'.format(func_or_iter))

    def get(self):
        """
        Get a 2 element tuple of this bag's args and kwargs.

        :return: tuple
        """
        return self.args, self.kwargs

    def extend(self, *args, **kwargs):
        return type(self)(*args, _parent=self, **kwargs)

    def set_parent(self, parent):
        self._parent = parent

    @classmethod
    def inherit(cls, *args, **kwargs):
        return cls(*args, _flags=(INHERIT_INPUT, ), **kwargs)

    def __eq__(self, other):
        return isinstance(other, Bag) and other.args == self.args and other.kwargs == self.kwargs

    def __repr__(self):
        return '<{} ({})>'.format(
            type(self).__name__, ', '.join(
                itertools.chain(
                    map(repr, self.args),
                    ('{}={}'.format(k, repr(v)) for k, v in self.kwargs.items()),
                )
            )
        )


class LoopbackBag(Bag):
    default_flags = (LOOPBACK, )


class ErrorBag(Bag):
    pass



import pytest

import bonobo
from bonobo.execution import GraphExecutionContext
from unittest.mock import patch


@pytest.mark.timeout(2)
def test_run_graph_noop():
    graph = bonobo.Graph(bonobo.noop)
    assert len(graph) == 1

    with patch('bonobo._api._is_interactive_console', side_effect=lambda: False):
        result = bonobo.run(graph)
    assert isinstance(result, GraphExecutionContext)



import csv

from bonobo.config import Option
from bonobo.config.processors import ContextProcessor
from bonobo.constants import NOT_MODIFIED
from bonobo.nodes.io.file import FileReader, FileWriter
from bonobo.nodes.io.base import FileHandler, IOFormatEnabled
from bonobo.util.objects import ValueHolder


class CsvHandler(FileHandler):
    """

    .. attribute:: delimiter

        The CSV delimiter.

    .. attribute:: quotechar

        The CSV quote character.

    .. attribute:: headers

        The list of column names, if the CSV does not contain it as its first line.

    """
    delimiter = Option(str, default=';')
    quotechar = Option(str, default='"')
    headers = Option(tuple, required=False)


class CsvReader(IOFormatEnabled, FileReader, CsvHandler):
    """
    Reads a CSV and yield the values as dicts.

    .. attribute:: skip

        The amount of lines to skip before it actually yield output.

    """

    skip = Option(int, default=0)

    @ContextProcessor
    def csv_headers(self, context, fs, file):
        yield ValueHolder(self.headers)

    def read(self, fs, file, headers):
        reader = csv.reader(file, delimiter=self.delimiter, quotechar=self.quotechar)

        if not headers.get():
            headers.set(next(reader))
        _headers = headers.get()

        field_count = len(headers)

        if self.skip and self.skip > 0:
            for _ in range(0, self.skip):
                next(reader)

        for row in reader:
            if len(row) != field_count:
                raise ValueError('Got a line with %d fields, expecting %d.' % (
                    len(row),
                    field_count,
                ))

            yield self.get_output(dict(zip(_headers, row)))


class CsvWriter(IOFormatEnabled, FileWriter, CsvHandler):
    @ContextProcessor
    def writer(self, context, fs, file, lineno):
        writer = csv.writer(file, delimiter=self.delimiter, quotechar=self.quotechar, lineterminator=self.eol)
        headers = ValueHolder(list(self.headers) if self.headers else None)
        yield writer, headers

    def write(self, fs, file, lineno, writer, headers, *args, **kwargs):
        row = self.get_input(*args, **kwargs)
        if not lineno:
            headers.set(headers.value or row.keys())
            writer.writerow(headers.get())
        writer.writerow(row[header] for header in headers.get())
        lineno += 1
        return NOT_MODIFIED



from unittest.mock import MagicMock

import pytest

import bonobo
from bonobo.config.processors import ContextCurrifier
from bonobo.constants import NOT_MODIFIED


def test_count():
    with pytest.raises(TypeError):
        bonobo.count()

    context = MagicMock()

    with ContextCurrifier(bonobo.count).as_contextmanager(context) as stack:
        for i in range(42):
            stack()

    assert len(context.method_calls) == 1
    bag = context.send.call_args[0][0]
    assert isinstance(bag, bonobo.Bag)
    assert 0 == len(bag.kwargs)
    assert 1 == len(bag.args)
    assert bag.args[0] == 42


def test_identity():
    assert bonobo.identity(42) == 42


def test_limit():
    context, results = MagicMock(), []

    with ContextCurrifier(bonobo.Limit(2)).as_contextmanager(context) as stack:
        for i in range(42):
            results += list(stack())

    assert results == [NOT_MODIFIED] * 2


def test_limit_not_there():
    context, results = MagicMock(), []

    with ContextCurrifier(bonobo.Limit(42)).as_contextmanager(context) as stack:
        for i in range(10):
            results += list(stack())

    assert results == [NOT_MODIFIED] * 10


def test_limit_default():
    context, results = MagicMock(), []

    with ContextCurrifier(bonobo.Limit()).as_contextmanager(context) as stack:
        for i in range(20):
            results += list(stack())

    assert results == [NOT_MODIFIED] * 10


def test_tee():
    inner = MagicMock(side_effect=bonobo.identity)
    tee = bonobo.Tee(inner)
    results = []
    for i in range(10):
        results.append(tee('foo'))

    assert results == [NOT_MODIFIED] * 10
    assert len(inner.mock_calls) == 10


def test_noop():
    assert bonobo.noop(1, 2, 3, 4, foo='bar') == NOT_MODIFIED



import io
import sys
from contextlib import redirect_stdout

from colorama import Style, Fore, init

init(wrap=True)

from bonobo import settings
from bonobo.plugins import Plugin
from bonobo.util.term import CLEAR_EOL, MOVE_CURSOR_UP


class IOBuffer():
    """
    The role of IOBuffer is to overcome the problem of multiple threads wanting to write to stdout at the same time. It
    works a bit like a videogame: there are two buffers, one that is used to write, and one which is used to read from.
    On each cycle, we swap the buffers, and the console plugin handle output of the one which is not anymore "active".

    """

    def __init__(self):
        self.current = io.StringIO()
        self.write = self.current.write

    def switch(self):
        previous = self.current
        self.current = io.StringIO()
        self.write = self.current.write
        try:
            return previous.getvalue()
        finally:
            previous.close()

    def flush(self):
        self.current.flush()


class ConsoleOutputPlugin(Plugin):
    """
    Outputs status information to the connected stdout. Can be a TTY, with or without support for colors/cursor
    movements, or a non tty (pipe, file, ...). The features are adapted to terminal capabilities.

    On Windows, we'll play a bit differently because we don't know how to manipulate cursor position. We'll only
    display stats at the very end, and there won't be this "buffering" logic we need to display both stats and stdout.

    .. attribute:: prefix

        String prefix of output lines.

    """

    def initialize(self):
        self.prefix = ''
        self.counter = 0
        self._append_cache = ''
        self.isatty = sys.stdout.isatty()
        self.iswindows = (sys.platform == 'win32')

        self._stdout = sys.stdout
        self.stdout = IOBuffer()
        self.redirect_stdout = redirect_stdout(self._stdout if self.iswindows else self.stdout)
        self.redirect_stdout.__enter__()

    def run(self):
        if self.isatty and not self.iswindows:
            self._write(self.context.parent, rewind=True)
        else:
            pass  # not a tty, or windows, so we'll ignore stats output

    def finalize(self):
        self._write(self.context.parent, rewind=False)
        self.redirect_stdout.__exit__(None, None, None)

    def write(self, context, prefix='', rewind=True, append=None):
        t_cnt = len(context)

        if not self.iswindows:
            buffered = self.stdout.switch()
            for line in buffered.split('\n')[:-1]:
                print(line + CLEAR_EOL, file=sys.stderr)

        alive_color = Style.BRIGHT
        dead_color = Style.BRIGHT + Fore.BLACK

        for i in context.graph.topologically_sorted_indexes:
            node = context[i]
            name_suffix = '({})'.format(i) if settings.DEBUG.get() else ''
            if node.alive:
                _line = ''.join(
                    (
                        ' ',
                        alive_color,
                        '+',
                        Style.RESET_ALL,
                        ' ',
                        node.name,
                        name_suffix,
                        ' ',
                        node.get_statistics_as_string(),
                        Style.RESET_ALL,
                        ' ',
                    )
                )
            else:
                _line = ''.join(
                    (
                        ' ',
                        dead_color,
                        '-',
                        ' ',
                        node.name,
                        name_suffix,
                        ' ',
                        node.get_statistics_as_string(),
                        Style.RESET_ALL,
                        ' ',
                    )
                )
            print(prefix + _line + '\033[0K', file=sys.stderr)

        if append:
            # todo handle multiline
            print(
                ''.join(
                    (
                        ' `-> ', ' '.join('{}{}{}: {}'.format(Style.BRIGHT, k, Style.RESET_ALL, v) for k, v in append),
                        CLEAR_EOL
                    )
                ),
                file=sys.stderr
            )
            t_cnt += 1

        if rewind:
            print(CLEAR_EOL, file=sys.stderr)
            print(MOVE_CURSOR_UP(t_cnt + 2), file=sys.stderr)

    def _write(self, graph_context, rewind):
        if settings.PROFILE.get():
            if self.counter % 10 and self._append_cache:
                append = self._append_cache
            else:
                self._append_cache = append = (
                    ('Memory', '{0:.2f} Mb'.format(memory_usage())),
                    # ('Total time', '{0} s'.format(execution_time(harness))),
                )
        else:
            append = ()
        self.write(graph_context, prefix=self.prefix, append=append, rewind=rewind)
        self.counter += 1


def memory_usage():
    import os, psutil
    process = psutil.Process(os.getpid())
    return process.memory_info()[0] / float(2**20)



def require(package, requirement=None):
    requirement = requirement or package

    try:
        return __import__(package)
    except ImportError:
        from colorama import Fore, Style
        print(
            Fore.YELLOW,
            'This example requires the {!r} package. Install it using:'.
            format(requirement),
            Style.RESET_ALL,
            sep=''
        )
        print()
        print(
            Fore.YELLOW,
            '  $ pip install {!s}'.format(requirement),
            Style.RESET_ALL,
            sep=''
        )
        print()
        raise



# Bonobo data-processing toolkit.
#
# Bonobo is a line-by-line data-processing toolkit for python 3.5+ emphasizing simplicity and atomicity of data
# transformations using a simple directed graph of python callables.
#
# Licensed under Apache License 2.0, read the LICENSE file in the root of the source tree.
"""Bonobo data-processing toolkit main module."""

import sys

assert (sys.version_info >= (3, 5)), 'Python 3.5+ is required to use Bonobo.'
from bonobo._api import *
from bonobo._api import __all__
from bonobo._version import __version__

__all__ = ['__version__'] + __all__
__version__ = __version__

del sys



import json

from bonobo.commands.run import read, register_generic_run_arguments
from bonobo.constants import BEGIN
from bonobo.util.objects import get_name

OUTPUT_GRAPHVIZ = 'graphviz'


def _ident(graph, i):
    escaped_index = str(i)
    escaped_name = json.dumps(get_name(graph[i]))
    return '{{{} [label={}]}}'.format(escaped_index, escaped_name)


def execute(*, output, **kwargs):
    graph, plugins, services = read(**kwargs)

    if output == OUTPUT_GRAPHVIZ:
        print('digraph {')
        print('  rankdir = LR;')
        print('  "BEGIN" [shape="point"];')

        for i in graph.outputs_of(BEGIN):
            print('  "BEGIN" -> ' + _ident(graph, i) + ';')

        for ix in graph.topologically_sorted_indexes:
            for iy in graph.outputs_of(ix):
                print('  {} -> {};'.format(_ident(graph, ix), _ident(graph, iy)))

        print('}')
    else:
        raise NotImplementedError('Output type not implemented.')


def register(parser):
    register_generic_run_arguments(parser)
    parser.add_argument('--graph', '-g', dest='output', action='store_const', const=OUTPUT_GRAPHVIZ)
    parser.set_defaults(output=OUTPUT_GRAPHVIZ)
    return execute



import os

import bonobo


def extract():
    env_test_user = os.getenv('ENV_TEST_USER')
    env_test_number = os.getenv('ENV_TEST_NUMBER')
    env_test_string = os.getenv('ENV_TEST_STRING')
    return env_test_user, env_test_number, env_test_string


def load(s: str):
    print(s)


graph = bonobo.Graph(extract, load)

if __name__ == '__main__':
    bonobo.run(graph)



from bonobo.execution.base import LoopingExecutionContext, recoverable


class PluginExecutionContext(LoopingExecutionContext):
    PERIOD = 0.5

    def __init__(self, wrapped, parent):
        # Instanciate plugin. This is not yet considered stable, as at some point we may need a way to configure
        # plugins, for example if it depends on an external service.
        super().__init__(wrapped(self), parent)

    def start(self):
        super().start()

        with recoverable(self.handle_error):
            self.wrapped.initialize()

    def shutdown(self):
        if self.started:
            with recoverable(self.handle_error):
                self.wrapped.finalize()
        self.alive = False

    def step(self):
        with recoverable(self.handle_error):
            self.wrapped.run()
